# -*- coding: utf-8 -*-
"""
Created on Fri Apr  7 03:15:02 2017

@author: Yuki
"""
import sys
import pickle
import inspect
from PyQt5.QtCore import QRect,pyqtSignal,Qt
from PyQt5.QtWidgets import QFileDialog,QWidget,QPushButton,QSizePolicy,QHBoxLayout,QVBoxLayout,\
                            QTreeView,QAbstractItemView,QApplication,QMainWindow,QToolBar,QAction,QToolButton,QHeaderView,QTextEdit
from PyQt5.QtGui import QColor,QStandardItem,QStandardItemModel,QFont

import coatl.Tree as MyTree
from coatl.Graph import MyGraphWindow

NAME=0
TYPE=1

def getRoot():
    path = QFileDialog.getSaveFileName(None, 'choose a save file','root.jh','JH file (*.jh)',options=QFileDialog.DontConfirmOverwrite)[0] #pyqt5ではtapleの一個めにpathが入っている
    if not path=='': #when the user appointed a file
        import os
        if os.path.isfile(path): #when file exists
            try:
                root=pickle.load(open(path,'rb')) #when can unpickle
                root.setview.setSavePath(path)
            except Exception as e:
                print(e)
                try:
                    depend_file=path+'_depend' #when can find a dependencies file
                    dependencies=pickle.load(open(depend_file,'rb'))
                    print('This file depends on the packages below. Check your environment meets these requirements.')
                    print(dependencies)
                except Exception as e2:
                    print(e2)
                    print('Cannot find the dependencies file. This file seems to depend on packages whick do not match your environment')
                root=MyTreeWidget()
                root.setview.setSavePath(path)
        else: #when file does not exist
            f=open(path,'wb')
            root=MyTreeWidget()
            root.setview.setSavePath(path)
            
        return root

def geneMyTreeWidget(view):
    return MyTreeWidget(view=view)
        
class MyTreeWidget(QMainWindow):
    def __init__(self,view=None):
        super().__init__()
        if view==None:
            self.setview=MyTreeView()
        else:
            self.setview=view
        self.cur=self.setview.cur
        self.root=self.setview.setmodel.mytree
        self.setview.setmodel.mytree.currentSignal.connect(self.setCurrent)
        
        self.initUI()
        
    def initUI(self):
        button_style='''
                    QPushButton {background-color: rgb(39,105,195);
                                 border-style: none;
                                 border-radius: 5px;
                                 font: bold 18px;
                                 min-width: 5em;
                                 padding: 6px;
                                 }
                    QPushButton:pressed {background-color: rgb(200,200,200);}
                    '''
        addButton=QPushButton('+')
        delButton=QPushButton('-')
        addButton.setStyleSheet(button_style)
        delButton.setStyleSheet(button_style)
        addButton.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        delButton.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        addButton.pressed.connect(self.setview.add)
        delButton.pressed.connect(self.setview.delete)
        bbox=QHBoxLayout()
        bbox.addWidget(addButton)
        bbox.addWidget(delButton)
        bbox.addStretch(1)
        
        self.vbox=QVBoxLayout()
        self.vbox.addLayout(bbox)
        self.vbox.addWidget(self.setview)
        
        self.centralWidget=QWidget()
        self.centralWidget.setLayout(self.vbox)
        self.setCentralWidget(self.centralWidget)
        self.setGeometry(QRect(1, 45, 499, 974))
        self.setWindowTitle('MyTreeWidget')
        self.show()
        
    def __reduce_ex__(self,proto):
        return geneMyTreeWidget,(self.setview,)
        
    def save(self,other=False):
        if other: #self.filepath以外の保存先を指定する場合
            path = QFileDialog.getSaveFileName(None, 'chose save file','root.jh')[0]
            if not path=='': #pyqt5ではtapleの一個めにpathが入っている
                depend_path=self.setview.filepath+'_depend'
                pickle.dump(self,open(path,'wb'))
                pickle.dump(self.get_dependencies(),open(depend_path,'wb'))
                print('Saved To::')
                print(path)
        else: #default 
            pickle.dump(self,open(self.setview.filepath,'wb'))
            depend_path=self.setview.filepath+'_depend'
            pickle.dump(self.get_dependencies(),open(depend_path,'wb'))
            print('Saved To::')
            print(self.setview.filepath)
            
    def refresh(self):
        self.setview.resetView()
        self.cur=self.setview.cur
        self.setview.setmodel.mytree.currentSignal.connect(self.setCurrent)
        
    def setCurrent(self,mylist):
        self.cur=mylist[0]

    def get_dependencies(self):
        '''
        Return the versions of packages whose classes are contained in self
        
        Returns:
            dict
        '''
        
        return self.root.get_dependencies()
        
    def do_same(self,function):
        '''move the self.cur to its children and execute the function respectively'''
        child_folders=[child for child in self.cur.getChildren().values() if isinstance(child,MyTree.MyTree)]
        for child in child_folders:
            self.root.setCurrentDirectly(child)
            function(self.cur)
            
    def set_current(self,mylist):
        '''set the current folder by a full path. (Example):['root','folder1','folder1-1']'''
        self.root.setCurrent(mylist)
        
def geneMyTreeView(model):
    return MyTreeView(model=model)

class MyTreeView(QTreeView):
    def __init__(self,model=None):
        super().__init__()
        self.setDragDropMode(QAbstractItemView.InternalMove)
        self.setSortingEnabled( True )

        self.expanded.connect(self.showGraphs)
        self.collapsed.connect(self.showGraphs)
        if model==None:
            self.setmodel=MyTreeModel(MyTree.MyRootTree())
        else:
            self.setmodel=model
        self.cur=self.setmodel.mytree.current #mytreeのcurrentが常に入っている
        self.setmodel.mytree.currentSignal.connect(self.setCurrent)
        self.setModel(self.setmodel)
        self.doubleClicked.connect(self.setmodel.double_clicked)
        self.expandAll()
        self.setGeometry(QRect(1, 45, 496, 974))
        self.show()
        
    def __reduce_ex__(self,proto):
        #.dumpファイルは移動されるかもしれないのでpickle時にself.filepathは保存しない
        return geneMyTreeView,(self.model(),)
        
    def add(self,index=0):
        if 'folder'+str(index) in self.cur.getChildren().keys(): #同じ名前がある場合
            index+=1
            self.add(index)
        else:
            self.cur.add(MyTree.MyTree(),'folder'+str(index))
    
    def delete(self):
        selected=self.selectedIndexes()
        if len(selected)>0:
            index = self.selectedIndexes()[0]
            item=self.model().itemFromIndex(index)
            path=self.model().ascend(item)
            if len(path)>1: #['root']は消さない
                label=path.pop(-1)
                self.model().mytree.dele_this(path,label,signal=True)
        
    def setSavePath(self,filepath):
        self.filepath=filepath
        
    def setCurrent(self,mylist):
        self.cur=mylist[0]

    def resetView(self):
        #Viewとmytreeの間にずれが生じたときにmytreeに合わせてViewを初期化
        mytree=self.model().mytree
        self.setmodel=MyTreeModel(mytree)
        self.cur=self.setmodel.mytree.current #mytreeのcurrentが常に入っている
        self.setmodel.mytree.currentSignal.connect(self.setCurrent)
        self.setModel(self.setmodel)
        self.doubleClicked.connect(self.setmodel.double_clicked)
        return self
        
    def showGraphs(self,index):
        #表示するべきグラフの参照のリストを返す
        model=self.model()
        root=model.getRoot()
        open_graphs=self.getOpenGraphs(root,model)
        close_graphs=self.getCloseGraphs(root,model)
        for g in open_graphs:
            g.show()
        for g in close_graphs:
            g.hide()
        
    def getOpenGraphs(self,item,model):
        #item以下にある表示するべきグラフを返すcollapseされているフォルダは調べない
        if self.isExpanded(model.indexFromItem(item))==True:
            graphs=model.getGraphs(item) #直下のグラフ
            childItems=self.getChildren(item)
            for child in childItems:
                ref=model.getRef(child)
                if isinstance(ref,MyTree.MyTree):
                    graphs=graphs+self.getOpenGraphs(child,model)
            return graphs
        else:
            return []

    def getCloseGraphs(self,item,model):
        #item以下にある隠すべきグラフのリストを返す
        if self.isExpanded(model.indexFromItem(item))==True:
            graphs=[]
            childItems=self.getChildren(item)
            for child in childItems:
                ref=model.getRef(child)
                if isinstance(ref,MyTree.MyTree):
                    graphs=graphs+self.getCloseGraphs(child,model)
            return graphs
        else:
            return  self.getAllGraphs(item,model)
                    
    def getAllGraphs(self,item,model):
        #item以下にある全てのグラフのリストを返す
        graphs=model.getGraphs(item) #直下のグラフ
        childItems=self.getChildren(item)
        for child in childItems:
            ref=model.getRef(child)
            if isinstance(ref,MyTree.MyTree):
                graphs=graphs+self.getAllGraphs(child,model)
        return graphs
        
    @classmethod
    def getChildren(cls,item):
        childItems=[]
        for i in range(item.rowCount()):
            childItems.append(item.child(i,0))
        return childItems
        
    def save(self,other=False):
        if other: #self.filepath以外の保存先を指定する場合
            path = QFileDialog.getSaveFileName(None, 'chose save file','root.jh')[0]
            if not path=='': #pyqt5ではtapleの一個めにpathが入っている
                pickle.dump(self,open(path,'wb'))
                print('Saved To::')
                print(path)
        else: #default 
            pickle.dump(self,open(self.filepath,'wb'))
            print('Saved To::')
            print(self.filepath)

def geneMyTreeModel(tree):
    return MyTreeModel(tree)
        
class MyTreeModel(QStandardItemModel):
    deleSignal=pyqtSignal(list,str)
    white=QColor(255,255,255)
    pink=QColor(197,133,217)
    red=QColor(255,0,0)
    black=QColor(0,0,0)
    green=QColor(10,210,10)
    def __init__(self,mytree):
        super().__init__(0,2)
        self.mytree=mytree #the instance of MyTree.MyRootTree
        self.mytree.addSignal.connect(self.add_to_model)
        self.mytree.deleSignal.connect(self.dele_from_model)
        self.mytree.renameSignal.connect(self.rename_model)
        self.mytree.currentSignal.connect(self.highlight_folder)
        
        self.setHeaderData( NAME,Qt.Horizontal, 'Name' )
        self.setHeaderData( TYPE,Qt.Horizontal, 'type' )
        self.containor=self.invisibleRootItem()
        self.containor.setDropEnabled(False)
        self.containor.appendRow(self.convert(mytree))
        
        #drag,dropに対応するために繋ぐが、mytreeからの操作があった時だけ一時的に切断する(mytreeに操作が反射するのを防ぐため)
        self.signalConnect()
        #GUI上のドラッグアンドドロップをmytreeに反映させるのに使う
        self.delePath=None
        self.deleLabel=None
        self.addPath=None
        self.addLabel=None
        
        #currentフォルダの背景色を変えるのに使う
        self.current=[]
        #currentフォルダへのアクセスに使う
        
        self.deleSignal.connect(self.dele_from_tree)
        
    def __reduce_ex__(self,proto):
        return geneMyTreeModel,(self.mytree,)
        
    def signalConnect(self):
        self.rowsRemoved.connect(self.remove, type=Qt.QueuedConnection)
        self.itemChanged.connect(self.changed)
        
    def signalDisconnect(self):
        self.rowsRemoved.disconnect(self.remove)
        self.itemChanged.disconnect(self.changed)
        
    def double_clicked(self,index):
        #TYPE上でダブルクリックされた時の動作を記述
        item=self.itemFromIndex(index)
        if item.column()==TYPE:
            ref=self.getRef(item)
            if isinstance(ref,MyTree.MyTree):
                #self.currentを更新してself.mytreeに反映　新旧のitemに通知
                self.mytree.setCurrentDirectly(ref)
            
            elif isinstance(ref,int):
                print(ref)
                
    def highlight_folder(self,mylist):
        '''Slot for currentSignal of MyRootTree. mylist is a list only contains the new current folder'''
        #dehighlighten the old current row
        pre=[]
        while len(self.current)>0:
            pre.append(self.current.pop())
        for i in pre:
            try:
                self.dataChanged.emit(self.indexFromItem(i),self.indexFromItem(i))
            except:
                print('current folder was deleted.')
            
        #get a new row to highligten
        cur=mylist[0]
        full_path=cur.ascend()
        target=self.pathToItem(full_path)
        self.current=self.getPaintRow(target)
        
        #highlighten the row
        for i in self.current:
            self.dataChanged.emit(self.indexFromItem(i),self.indexFromItem(i))
            
    def data(self, index, role =Qt.DisplayRole):
        #overriding method:indexで指定される要素に関する情報を返す
        if not index.isValid():
            return None
        else:
            item=self.itemFromIndex(index)
            
        if role ==Qt.DisplayRole:
            return item.text()
        elif role ==Qt.EditRole:
            return item.text()
        elif role ==Qt.BackgroundRole:
            if item in self.current:
                return self.pink
        elif role==Qt.ForegroundRole:
            if item in self.current:
                return self.white
            elif item.text()=='folder':
                return self.red
            elif item.column()==1:
                return self.green
            else:
                return self.black
        else:
            return None
        
    def setData(self, index, value, role = Qt.EditRole):
        #overriding method:viewから編集された時の挙動を記述する
        if not index.isValid():
            return False
        item=self.itemFromIndex(index)
        if item.column()==NAME:
            full_label=self.ascend(item)
            if role ==Qt.EditRole and value != "":
                if item.parent()==None: #rootは変更不可
                    return False
                else:
                    try:
                        if self.mytree.rename_this(full_label,value):
                            self.mytree.rename_this(full_label,value)
                            item.setText(value)
                            return True
                        else:
                            return False
                    except:
                        return False
            else:
                return False
        else:
            return False
            
    def headerData(self,section,orientation,role=Qt.EditRole):
        if role==Qt.ForegroundRole:
            result=self.black
        else:
            result=super().headerData(section,orientation,role)
        return result
            
    def getRef(self,item):
        #itemの行のNAMEに対応するmytreeの参照を返す
        if item.parent()==None: #rootフォルダの場合
            return self.mytree
        else: 
            parent=item.parent()
            num=item.row()
            name=parent.child(num,NAME)
            full_path=self.ascend(name)
            return self.mytree.get_this(full_path)
            
    def getPaintRow(self,item):
        #カレントディレクトリとして色を変えて表示するitemのリストを返す
        if item.parent()==None: #rootフォルダの場合
            return [self.containor.child(0,i) for i in range(self.containor.columnCount())]
        else:
            parent=item.parent()
            num=item.row()
            return [parent.child(num,i) for i in range(parent.columnCount())]
          
    def convert(self,mytree):
        if not isinstance(mytree,MyTree.MyTree):
            print(mytree)
            raise TypeError('MyTree以外の変換は実装していません')
        else:
            folder=MyItemList(mytree.name,'folder')
            children=mytree.getChildren()
            for key,item in children.items():
                if isinstance(item,MyTree.MyTree): #子供がフォルダーの場合
                    folder.addChild(self.convert(item))
                else: #子供がデータだった場合
                    data=MyItemList(key,item.__class__.__name__)
                    data[NAME].setDropEnabled(False)
                    folder.addChild(data)
            return folder
        
    def remove(self, index, first, last):
        #indexには除かれたアイテムの親のindexがはいる
        if index.isValid():
            parent = self.itemFromIndex(index)
            path=self.ascend(parent)
        self.delePath=path
        if not self.deleLabel==None:
            self.deleSignal.emit(self.delePath,self.deleLabel)
        
    def changed(self,item):
        #行が追加された時以外にも呼ばれるのでエラーの原因になるかも　今のところは大丈夫そう
        if item.column()==NAME and not item.parent()==None:
            self.deleLabel=item.text()
            self.addLabel=item.text()
            self.addPath=self.ascend(item.parent())
            
            if not self.delePath==None:
                self.deleSignal.emit(self.delePath,self.deleLabel)
                
    def dele_from_tree(self,path,label):
        addRef=self.mytree.dele_this(path,label)
        self.mytree.add_this(self.addPath,self.addLabel,addRef)
        self.delePath=None
        self.deleLabel=None
        
    def add_to_model(self,path,label,ref):
        #mytreeに追加された要素をmodelにも反映させる　refは長さ1のリスト mytreeへ重複した操作を反射しないようにするために一時的にシグナルを切る
        self.signalDisconnect()
        parent=self.pathToItem(path)
        target=ref[0]
        if isinstance(target,MyTree.MyTree): #子供がフォルダーの場合
            parent.appendRow(self.convert(target))
        else: #子供がデータだった場合
            data=MyItemList(label,target.__class__.__name__)
            data[NAME].setDropEnabled(False)
            parent.appendRow(data)
        self.dataChanged.emit(self.indexFromItem(parent),self.indexFromItem(parent))
        self.signalConnect()
            
    def dele_from_model(self,path,label):
        #mytreeから削除された要素をmodelからも削除する
        self.signalDisconnect()
        parent=self.pathToItem(path)
        for i in range(parent.rowCount()):
            if parent.child(i,0).text()==label:
                parent.removeRow(i)
                break
        self.dataChanged.emit(self.indexFromItem(parent),self.indexFromItem(parent))
        self.signalConnect()
            
    def rename_model(self,path,before,after):
        #mytreeで改名された要素に対応するmodelの要素も改名する
        self.signalDisconnect()
        parent=self.pathToItem(path)
        for i in range(parent.rowCount()):
            if parent.child(i,0).text()==before:
                parent.child(i,0).setText(after)
                break
        self.dataChanged.emit(self.indexFromItem(parent),self.indexFromItem(parent))
        self.signalConnect()
        
    def pathToItem(self,path):
        #pathに対応するitemを返す
        start=self.getRoot()
        if path==['root']:
            return start
        else:
            for child in path[1:]:
                for i in range(start.rowCount()):
                    if start.child(i,0).text()==child:
                        start=start.child(i,0)
                        break
                    if i==start.rowCount()-1: #見つからない場合
                        raise Exception('pathで指定されたitemが見つかりません')
            return start
        
            
    def ascend(self,start):
        #startはQStandardItem invisibleRootからのパスを['root',....]のように返す
        path=[start.text()]
        while start.parent():
            start=start.parent()
            path.append(start.text())
        path.reverse()
        return path
        
    def getGraphs(self,item):
        #自身の直下（フォルダは展開しない）のグラフの参照のリストを返す
        ref=self.getRef(item)
        if isinstance(ref,MyTree.MyTree):
            children=ref.getChildren()
            graphs=[]
            for child in children.values():
                if isinstance(child,MyGraphWindow):
                    graphs.append(child)
            return graphs
        else:
            return []
            
    def getRoot(self):
        #root itemを返す
        root=self.containor.child(0,0)
        return root
                 
class MyItemList(list):
    def __init__(self,*args):
        self.items=[QStandardItem(arg) for arg in args]
        for item in self.items[TYPE:]:
            item.setDropEnabled(False)
        self.items[TYPE].setEditable(False)
        super().__init__(self.items)
        self.children=[]

    def setDragEnabled(self,boolean):
        self.items[0].setDragEnabled(boolean)
        
    def addChild(self,itemlist):
        if isinstance(itemlist,MyItemList):
            self.items[0].appendRow(itemlist)
        else:
            raise TypeError('MyItemListしか追加できません')
            
if __name__=='__main__':
    app = QApplication( sys.argv )
    mw=MyTreeWidget()
    mytree=mw.cur
    folder1=MyTree.MyTree(children={'1_a':1.1,'1_b':'string','1_c':None})
    folder2=MyTree.MyTree(children={'2_a':1.1,'2_b':'string','2_c':None})
    folder3=MyTree.MyTree(children={'3_a':1.1,'3_b':'string','3_c':None})
    folder4=MyTree.MyTree(children={'4_a':1.1,'4_b':'string','4_c':None})
    folder3.add(folder4)
    folder2.add(folder3)
    folder1.add(folder2)
    mytree.add(folder1)
    sys.exit( app.exec_() )


# -*- coding: utf-8 -*-
"""
Created on Sat Oct  7 22:16:39 2017

@author: Yuki
"""

from PyQt5.QtWidgets import QWidget,QApplication,QMainWindow,QMdiArea,QMdiSubWindow,QMenuBar,QAction,QFileDialog,QDesktopWidget,\
                            QStatusBar
from PyQt5.QtCore import pyqtSignal                    
import pickle,os

import coatl
from coatl.Graph import MyGraphWindow
from coatl.Browser import MyTreeWidget
from coatl.Logger import logger

MAIN_EXTENSION='cl'
DEPEND_EXTENSION='cld'

MAXIMIZED=0
MINIMIZED=1
NORMAL=2

def open_windows():
    pathes=QFileDialog.getOpenFileNames(None,'Choose files to open','','{0} file (*{1})'.format(coatl.COATL,MAIN_EXTENSION))
    
    def try_pickle(path):
        if os.path.isfile(path):
            depend_path=path.rstrip(MAIN_EXTENSION)+DEPEND_EXTENSION
            try:
                answer=pickle.load(open(path,'rb'))
                answer.set_path(path,depend_path)
                return answer
            except Exception as e:
                print(e)
                try:
                    with open(depend_path,'rb') as df:
                        depend=pickle.load(df)
                        print('This {} file depends on the packages below. Check your environment meets these requirements.'.format(coatl.COATL))
                        print(depend)
                except Exception as e2:
                    print(e2)
                    print('Cannot find the dependencies file with extension (.{}). \
                          This file may depend on packages which do not match your environment'.format(DEPEND_EXTENSION))
        else:
            raise FileNotFoundError
            
    if len(pathes[0])==1:
        answer=try_pickle(pathes[0][0])
        return answer
    elif len(pathes[0])>1:
        answer=[try_pickle(path) for path in pathes[0]]
        return answer

class Manager():
    '''Manage widgets by holding them as attributes.'''
    
    
    def __init__(self,mdi,target_class):
        '''
        Params:
            mdi: QMdiArea
                the QMdiArea to which widgets are added
            target_class: class
                the class of widgets
        '''
        self.__mdi=mdi
        self.__target_class=target_class
        self.__prefix='_{}__'.format(self.__class__.__name__)
    
    def _add(self,widget,name,state=None,geometry=None):
        '''
        Add a widget to self.__mdi and holds it as an atrribute of itself
        
        Paramters:
            widget: QWidget
                the widget to be added (must be able to be pickled)
            name: str
                the widget's name
        '''
        if isinstance(widget,self.__target_class):
            subwindow=SubWindow()
            subwindow.setWidget(widget)
            subwindow.deleteSignal.connect(self._delete)
            self.__mdi.addSubWindow(subwindow)
            widget.objectNameChanged.connect(subwindow.setWindowTitle)
            checked=self._check_name(name)
            widget.setObjectName(checked)
            if state==MAXIMIZED:
                subwindow.showMaximized()
            elif state==MINIMIZED:
                subwindow.showMinimized()
            elif state==NORMAL:
                if not geometry==None:
                    subwindow.showNormal()
                    subwindow.setGeometry(geometry)
            widget.show()
            setattr(self,checked,widget)
        else:
            raise Exception('A widget to be added to this instance must be an instance of {}'.format(self.__target_class))
        
    def _delete(self,name):
        '''
        Delete a widget specified by the name from self.__mdi and the attributes
        
        Parameters:
            name: str
                the widget's name
        '''
        widgets=self._get_all_widgets()
        if name in widgets.keys():
            self.__mdi.removeSubWindow(widgets[name].parent())
            delattr(self,name)
        else:
            raise Exception('{} does not exist.'.format(name))
            
    def _get_all_widgets(self):
        '''
        Return All the widgets as dict
        
        Returns:
            dict
                key: the widget's name value: the widget
        '''
        answer={}
        for key,value in vars(self).items():
            if not key.startswith(self.__prefix):
                answer[key]=value
        return answer
        
    def _check_name(self,name):
        '''
        Check if 'name' already exists in self attributes' names  and return an appropriately suffixed name if needed.
        If the name starts with self.__prefix, raise Exception.
        
        Parameters:
            name: str
                the widget's name
        Returns:
            str
                an appropriate name suffixed if needed
        '''
        def check_and_get(name,mydict):
            '''
            Check the keys of 'mydict' and return an appropriate, suffixed if needed, string near 'name'.
            '''
            keys=mydict.keys()
            suffix=0
            answer=name
            while True:
                if not answer in keys:
                    return answer
                else:
                    answer=name+str(suffix)
                    suffix=suffix+1
            return answer
            
        if name.startswith(self.__prefix):
            raise Exception('\'{}\' is not an available name'.format(name))
        else:
            return check_and_get(name,self._get_all_widgets())
        
    def _rename(self,old,new):
        '''
        Rename a widget
        
        Params:
            old: str
                the old name
            new: str
                a new name
        '''
        widgets=self._get_all_widgets()
        if old in widgets.keys():
            widget=widgets[old]
            checked=self._check_name(new)
            delattr(self,old)
            setattr(self,checked,widget)
            widget.setObjectName(checked)
        else:
            raise Exception('{} does not exist.'.format(old))

class SubWindow(QMdiSubWindow):
    deleteSignal=pyqtSignal(str)
    
    def closeEvent(self, event):
        self.deleteSignal.emit(self.widget().objectName())
        event.accept()
            
def unpickle_mainwindow(main_geometry,g_geometries,b_geometries):
    '''
    Unpickle a MainWindow, keeping all the geometries when pickled
    
    Params:
        main_geometry: QRect
        g_geometries: dict {str: (MyGraphWindow,int,QRect)}
            {the name of widget: (the widget (can be pickled),the state,geometry of the parent subwindow)}
        b_geometries: dict {str: (MyTreeWidget,int,QRect)}
            same as above
    Returns:
        MainWindow
    '''
    
    main=MainWindow()
    
    def place_widgets(geometries,add_method):
        for key,value in geometries.items():
            widget=value[0]
            state=value[1]
            geometry=value[2]
            add_method(widget,key,state,geometry)
            
    place_widgets(g_geometries,main.add_graph)
    place_widgets(b_geometries,main.add_browser)
    
    return main
    
class MainWindow(QMainWindow):
    '''The main window of this application which contains subwindows'''
    def __init__(self):
        super().__init__()
        self.initUI()
        self.graphs=Manager(self.mdiArea,MyGraphWindow) #holds graph windows
        self.browsers=Manager(self.mdiArea,MyTreeWidget) #holds browser windows
        self._path=None
        self._depend_path=None
        
        #place the window to the left of the screen
        screen=QDesktopWidget().availableGeometry()
        height=int(0.9*screen.height())
        width=int(0.9*screen.width()/2)
        self.setGeometry(0,40,width,height)
        self.setWindowTitle(coatl.COATL)
        self.show()
        
    def __reduce_ex__(self,protocol):
        '''
        For pickle
        '''
        def get_states(manager):
            '''
            Get the states and geometris of widgets
            
            Params:
                manager: Manager
            Returns:
                dict {str: (QWidget,int,QRect)}
                    {the name of widget: (the widget (can be pickled),the state,geometry of the parent subwindow)}
            '''
            
            widgets=manager._get_all_widgets()
            geometries={}
            for key,widget in widgets.items():
                subwindow=widget.parent()
                if subwindow.isMaximized():
                    state=MAXIMIZED
                elif subwindow.isMinimized():
                    state=MINIMIZED
                else:
                    state=NORMAL
                geo=subwindow.geometry()
                geometries[key]=(widget,state,geo)
            return geometries
        
        graph_geometries=get_states(self.graphs)
        browser_geometries=get_states(self.browsers)
                
        return unpickle_mainwindow,(self.geometry(),graph_geometries,browser_geometries)

    def initUI(self):   
        #add Menubar
        menubar=MyMenuBar(self)
        self.setMenuBar(menubar)
        menubar.saveAction.triggered.connect(self.save)
        menubar.saveAsAction.triggered.connect(self.save_as)
        self.mdiArea=QMdiArea()
        self.setCentralWidget(self.mdiArea)
        self.status=QStatusBar(self)
        self.setStatusBar(self.status)
        
    def new_graph(self,name='Unnamed'):
        '''
        Create a new graph window and add it to self.graphs
        
        Parameters
            name: the name of the graph
        Returns
            MyGraphWindow
        '''
        graph=MyGraphWindow()
        self.graphs._add(graph,name)
        return graph
        
    def new_browser(self,name='Unnamed'):
        '''
        Create a new browser and add it to self.browsers
        
        Parameters
            name: the name of the browser
        Returns
            MyTreeWidget
        '''
        browser=MyTreeWidget()
        self.browsers._add(browser,name)
        return browser
    
    def add_graph(self,graph,name='Unnamed',state=None,geometry=None):
        '''
        Add a graph to self.graphs
        
        Parameters
            graph: MyGraphWindow
            name: str
        '''
        self.graphs._add(graph,name,state,geometry)
            
    def add_browser(self,browser,name='Unnamed',state=None,geometry=None):
        '''
        Add a browser to self.browsers
        
        Parameters
            graph: MyTreeWidget
            name: str
        '''
        self.browsers._add(browser,name,state,geometry)
    
    def set_active(self,graph):
        '''
            Set the active graph for 'graph'
        '''
        pass
    
    def get_active(self):
        '''
            Get the active graph
        '''
        
    def get_graphs(self):
        '''
            Get all the graphs which this instance has.
        '''
        return self.graphs
        
    def save(self):
        '''
        Save itself to the file in self._path if set
        '''
        if self._path:
            self.pickle_self(self._path,self._depend_path)
        else:
            self.save_as()
    
    def save_as(self):
        '''
        Accept a file path via a file dialog and try to save itself to it
        '''
        path = QFileDialog.getSaveFileName(None, 'Choose a save file','temp.{}'.format(MAIN_EXTENSION),
                                           '{0} file (*{1})'.format(coatl.COATL,MAIN_EXTENSION))[0] #pyqt5ではtapleの一個めにpathが入っている
        if path:
            #Enforce the extension
            if not path.endswith('.{}'.format(MAIN_EXTENSION)):
                path=path+'.{}'.format(MAIN_EXTENSION)
            depend_path=path.rstrip(MAIN_EXTENSION)+DEPEND_EXTENSION
            self.pickle_self(path,depend_path)
                    
    def pickle_self(self,path,depend_path):
        '''
        Actual pickling
        
        Params:
            path: str
                the path of the file to which itself is writen
            depend_path: str
                the path of the file to which the dependencies are writen
        '''
        with open(path,'wb') as f:
            with open(depend_path,'wb') as df:
                #! must call self.set_path before pickling because pathes are needed when pickling
                self.set_path(path,depend_path)
                pickle.dump(self,f)
                #get all the dependent packages' version as dict and pickle it
                versions=[browser.get_dependencies() for browser in self.browsers._get_all_widgets().values()]
                merged_version={}
                for version in versions:
                    merged_version.update(version)
                from pkg_resources import get_distribution
                merged_version[coatl.GRAPH_PACKAGE]=get_distribution(coatl.GRAPH_PACKAGE).version
                pickle.dump(merged_version,df)

                self.status.showMessage('Saved to: {}'.format(path))    
    
    def set_path(self,path,depend_path):
        '''
        Set the file path to which itself is pickled
        
        Params:
            path: str
            depend_path: str
                the path of the file to which the dependencies are writen
        '''
        self._path=path
        self._depend_path=depend_path
        self.setWindowTitle('{0}--{1}'.format(coatl.COATL,path))
        
class MyMenuBar(QMenuBar):
    '''Menubar'''
    def __init__(self,ref):
        super().__init__()
        self.saveAction=QAction('save', self)
        self.saveAction.setShortcut('Ctrl+S')
        self.saveAsAction = QAction('save as', self)
        self.saveAsAction.setShortcut('Ctrl+A')
        
        self.fileMenu =self.addMenu('File')
        self.fileMenu.addAction(self.saveAction)
        self.fileMenu.addAction(self.saveAsAction)
        
        self.helpAction = QAction('About this program', self)
        self.helpAction.setShortcut('Ctrl+H')
        helpMenu=self.addMenu('Help')
        helpMenu.addAction(self.helpAction)
        
#        self.setState(ref.state)
#        ref.stateSignal.connect(self.setState)
        
#    def setState(self,end):
#        '''状態の遷移はこの関数を通して行われる'''
#        if end==RUNNING:
#            self.loadAction.setEnabled(False)
#        else:
#            self.loadAction.setEnabled(True)

if __name__=='__main__':
    import sys,pickle
    app = QApplication( sys.argv )
#    m=MainWindow()
#    m.new_graph()
#    m.new_graph()
#    m.new_browser()
#    m.new_browser()
#    g=MyGraphWindow()
#    m.add_graph(g,state=MINIMIZED)
    m=MainWindow()
#    sys.exit(app.exec_())        


# -*- coding: utf-8 -*-
"""
Created on Mon Apr  3 01:08:17 2017

@author: Yuki
"""
import sys
import inspect
import pandas as pd

from PyQt5.QtCore import pyqtSignal,QObject
from PyQt5.QtWidgets import QApplication,QWidget,QFileDialog

from coatl.Graph import MyGraphWindow

def transformMyTree(mytree,parent=None):
    #MyTreeをMyTreeRawに変換する pickleのため
    raw=MyTreeRaw(parent=parent,name=mytree.name)
    for key,value in mytree.getChildren().items():
        if isinstance(value,MyTree):
            raw[key]=transformMyTree(value,raw)
        else:
            raw[key]=value
    return raw

def transformMyRootTree(mytree,parent=None):
    #MyTreeをMyTreeRawに変換する pickleのため
    raw=MyTreeRaw(parent=parent,name=mytree.name)
    for key,value in mytree.getChildren().items():
        if isinstance(value,MyTree):
            raw[key]=transformMyTree(value,raw)
        else:
            raw[key]=value
    return raw
    
def transformMyTreeRaw(tree,parent=None):
    #MyTreeRawをMyTreeに変換する unpickleのため
    result=MyTree(parent=parent,name=tree.name)
    for key,value in tree.items():
        if isinstance(value,MyTreeRaw):
            result.add(ref=transformMyTreeRaw(value,result),label=key)
        else:
            result.add(ref=value,label=key)
    return result
    
def transformMyRootTreeRaw(tree,current_path,dependencies):
    #MyTreeRawをMyRootTreeに変換する unpickleのため
    try:
        result=MyRootTree(name=tree.name)
        for key,value in tree.items():
            if isinstance(value,MyTreeRaw):
                result.add(ref=transformMyTreeRaw(value,result),label=key)
            else:
                result.add(ref=value,label=key)
        result.setCurrent(current_path)
        return result
    except Exception as e:
        print(e)
        print(dependencies)
        raise Exception('Cannot unpickle the file.You may use a different environment from the one used when pickling. Use a environment satisfies the above requirements')
    
class MyTreeRaw(dict):
    '''Tree構造を実装するクラス　MyTreeをpickleする時にこれに変換する'''
    def __init__(self,parent,myobject=None,name='temp'):
        if myobject==None:
            super().__init__({})
        else:
            super().__init__(myobject)
        self.parent=parent #nodeはroot以外必ず親を持つ
        self.name=name #子の名前はdict型のkey
#        
#class MyRootTreeRaw(dict):
#    '''Convert MyRootTree to a dictionary to avoid recursive error when pickling. Also record package dependencies to show them up when the file is opened in a different environment'''
#    def __init__(self,parent,myobject=None,name='temp'):
#        if myobject==None:
#            super().__init__({})
#        else:
#            super().__init__(myobject)
#        self.parent=parent #nodeはroot以外必ず親を持つ
#        self.name=name #子の名前はdict型のkey
#        self.dependencies={}
    
class MyTree(QObject):
    '''
    A data folder class that has a tree structure. This class's instance holds children as its attributes and you can access them '.childname'. 
    Note that you must reimplement getChildren method when you add a new attribute to this class for the above reason.
    '''
    SPACE='  '
    INDENT='--'
    addSignal=pyqtSignal(list,str,list) #path,label,refの順 最後は参照を入れたいので[ref]とする
    deleSignal=pyqtSignal(list,str)
    renameSignal=pyqtSignal(list,str,str)
    
    def __init__(self,name='temp',parent=None,children=None):
        #childrenには子をdict型{'名前':参照}で渡す
        super().__init__()
        self.parent=parent #nodeはroot以外必ず親を持つ
        self.name=name
        if not children==None:
            for key,item in children.items():
                self.__dict__[key]=item
        
    def __reduce_ex__(self, proto):
        #pickleのためのメソッド 動的にインスタンス変数を追加するクラスはそのままpickleできない
        return transformMyTreeRaw,(transformMyTree(self),)
    
    def __str__(self, level=0,current=None,unfold=True):
        if self is current:
            ret = self.SPACE*level+self.INDENT+self.name+'<=='+"\n"
        else:
            ret = self.SPACE*level+self.INDENT+self.name+"\n"
        #MyTreeは先に展開それ以外のデータはunfoldがTrueならばkeyをprint   
        for key in sorted([key for key,value in self.getChildren().items() if isinstance(value,MyTree)]):
            ret += self.__dict__[key].__str__(level=level+1,current=current,unfold=unfold)
        if unfold:
            for key in sorted([key for key,value in self.getChildren().items() if not isinstance(value,MyTree)]):
                ret += self.SPACE*(level+1)+repr(key)+"\n"
        return ret
    
    def show(self,unfold=True):
        sys.stdout.write(self.__str__(unfold=unfold))
        
    def get(self,label):
        return self.__dict__[label]
    
    def getChildren(self):
        #子供を{'名前':参照}で返す
        children={k:v for k,v in self.__dict__.items() if not (k=='parent' or k=='name')}
        return children
    
    def add(self,ref,label=None,check=False,signal=True):
        #childrenにlabelと同一の名前が無ければchildとして加える checkは違う名前の同一オブジェクトが無いかチェックするオプション
        if label==None: #labelを指定していなければ呼び出し時の実引数をlabelとする
            frame = inspect.currentframe()
            stack = inspect.getouterframes(frame)
            val_name = stack[1].code_context[0].split('(')[1].split(')')[0] #これで実引数の名前を取得できるらしい ただし関数内やJupyterのcell内で連続してaddを呼び出すと最後のaddの実引数をlabelにするのでlabel重複のエラーがでる
            label=val_name
            
        if label in self.__dict__.keys():
            raise Exception('The same name already exists.Or you should call \'add\' method like this \'add(refference,\"name\")\' .')
        if check:
            result=self.checkChildren(ref)
            if result[0]:
                raise Exception('same object is registered as '+result[1])
            else:
                #フォルダの場合は親子関係を設定し、つけるラベルとフォルダの名前を一致させる addSignalは親に上げるためにtransmitAddSignalにconnect
                if isinstance(ref,MyTree):
                    ref.name=label
                    ref.parent=self
                    ref.addSignal.connect(self.transmitAddSignal)
                    ref.deleSignal.connect(self.transmitDeleSignal)
                    ref.renameSignal.connect(self.transmitRenameSignal)
                    self.__dict__[label]=ref
                else:
                    self.__dict__[label]=ref
        else:
            if isinstance(ref,MyTree):
                ref.name=label
                ref.parent=self
                ref.addSignal.connect(self.transmitAddSignal)
                ref.deleSignal.connect(self.transmitDeleSignal)
                ref.renameSignal.connect(self.transmitRenameSignal)
                self.__dict__[label]=ref
            else:
                self.__dict__[label]=ref

        #signalをemit
        if signal:
            self.addSignal.emit([self.name],label,[ref])
            
    def pop(self,label,signal=True):
        target=self.__dict__.pop(label)
        if isinstance(target,MyTree):
            target.disconnect()
        if signal:
            self.deleSignal.emit([self.name],label)
        return target
        
    def plot(self,x,y,xlabel='X',ylabel='Y',title='No name'):
        g=MyGraphWindow()
        g.plot(x,y)
        ax=g.fig.get_axes()[0]
        ax.set_title(title)
        g.setWindowTitle(title)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        
        def try_add(graph,label):
            self.add(graph,label)
            
        try:
            try_add(g,title)
        except:
            suffix=0
            while True:
                try:
                    try_add(g,title+str(suffix))
                    break
                except:
                    suffix+=1
        
        
    def rename(self,before,after,signal=True):
        if (not before==after) and (not after in self.getChildren().keys()): #beforeとafterが違って afterが子供にいない時
            ref=self.get(before)
            self.pop(before,signal=False) #ちなみにpopとaddを逆にすると挙動が変になる　popとaddでは同じオブジェクトを扱うがpopでシグナルをdisconnectしていることに注意
            self.add(ref,label=after,signal=False)
            if signal:
                self.renameSignal.emit([self.name],before,after)
            return True
        else:
            return False
        
    def checkChildren(self,ref):
        #子としてrefを持っていないかcheckする ref:オブジェクト参照
        result=[False,None]
        for key,child in self.getChildren().items():
            if id(child)==id(ref):
                result[0]=True
                result[1]=key
                break
        return result
        
    def ascend(self):
        #親を遡って一番上からのfull_pathを返す
        start=self
        full_path=[start.name]
        while not start.parent==None:
            start=start.parent
            full_path.append(start.name)
        full_path.reverse()
        return full_path
        
    def search(self,target):
        #カレントディレクトリ移動のためのサーチなのでディレクトリだけ調べる target:文字列
        if self.name==target:
            return {'result':True,'path':[self.name]}
        for child in self.getChildren().values():
            if isinstance(child,MyTree): #ディレクトリだけ調べる
                answer=child.search(target)
                if answer['result']:
                    return {'result':True,'path':([self.name]+answer['path'])}
        #ここまでくれば探索は解無し
        return {'result':False,'path':None}
    
    def runAll(self):
        #tree内を全ての参照のlistを返す
        mylist=[]
        mylist.append(self)
        for child in self.getChildren().values():
            if isinstance(child,MyTree):
                mylist=mylist+child.runAll()
            else:
                mylist.append(child)
        return mylist
    
    def loadFiles(self):
        def load_a_file(path):
            if not path=='':
                #1行だけ読み込んでみてstrが入っていればheaderとして使用してrootにもその名前で登録　そうでなければ'data0','data1',,,,としてrootに登録
                reader=pd.read_csv(path,sep='\t',comment='#',header=None,chunksize=1)
                data=reader.get_chunk(1)
                ndata=None
                if type(data.ix[0,0])==str:
                    ndata=pd.read_csv(path,sep='\t',comment='#')
                else:
                    ndata=pd.read_csv(path,sep='\t',comment='#',header=None)
                    ndata.columns=['data'+str(x) for x in range(len(ndata.columns))]
                    #自分の下にフォルダを作ってそこにデータを追加
                name=path.split('/')[-1]
                childFolder=MyTree()
                self.add(childFolder,label=name)
                for index in ndata.columns:
                    childFolder.add(ndata[index],label=index)           
                    
        paths = QFileDialog.getOpenFileNames(None, 'choose data files to load')[0]
        for path in paths:
            load_a_file(path)
            
    def transmitAddSignal(self,path,label,ref):
        #childのaddSignalのパスの先頭に自分の名前を付けくわえてemit
        path.insert(0,self.name)
        self.addSignal.emit(path,label,ref)
        
    def transmitDeleSignal(self,path,label):
        #childのdeleSignalのパスの先頭に自分の名前を付けくわえてemit
        path.insert(0,self.name)
        self.deleSignal.emit(path,label)
        
    def transmitRenameSignal(self,path,before,after):
        #childのrenameSignalのパスの先頭に自分の名前を付けくわえてemit
        path.insert(0,self.name)
        self.renameSignal.emit(path,before,after)
              
class MyRootTree(MyTree):
    '''カレントディレクトリを持つRootTree インスタンス変数を追加してはいけない'''
    #self.parent,self.name,self.current以外が子供なのでgetChildrenやpickle用の関数も上書きする
    DATA_PATH='.\\raw_data\\'
    currentSignal=pyqtSignal(list)
    def __init__(self,name='root',children=None):
        super().__init__(parent=None,name=name,children=children)
        self.current=self
        
    def __str__(self,unfold=True):
        return super().__str__(level=0,current=self.current,unfold=unfold)
        
    
    def __reduce_ex__(self, proto):
        #pickleのためのメソッド 動的にインスタンス変数を追加するクラスはそのままpickleできない
        return transformMyRootTreeRaw,(transformMyTree(self),self.current.ascend(),self.get_dependencies())
    
    def getChildren(self):
        #子供を{'名前':参照}で返す self.currentも子供以外の要素に加わったのでオーバーライド
        children={k:v for k,v in self.__dict__.items() if not (k=='parent' or k=='name' or k=='current')}
        return children 
        
    def add_this(self,path,label,ref,signal=False):
        #フルパスでの追加 path:['root',...] label:フォルダ名またはデータ名 ref:参照
        target=self
        for index,child in enumerate(path):
            if index>0:
                target=target.__dict__[child]
        target.add(ref,label=label,signal=signal)
            
    def dele_this(self,path,label,signal=False):
        #フルパスでの消去 path:['root',...]  label:フォルダ名またはデータ名
        target=self
        for index,child in enumerate(path):
            if index>0:
                target=target.__dict__[child]
        return target.pop(label,signal=signal)
        
    def rename_this(self,full_label,newname,signal=False):
        if full_label==['root']: #rootのrename
            self.name=newname
            return True
        elif len(full_label)>1:
            prename=full_label[-1]
            folder=self.get_this(full_label[:-1])
            return folder.rename(prename,newname,signal=signal)
        else:
            raise Exception('Canot rename')
            
    def setCurrent(self,full_label):
        self.current=self.get_this(full_label)
        self.currentSignal.emit([self.current])
        
    def setCurrentDirectly(self,mytree):
        if isinstance(mytree,MyTree):
            self.current=mytree
            self.currentSignal.emit([self.current])
        else:
            raise Exception('the current folder must be a MyTree')

    def get_this(self,full_label):
        #full_labelで指定されたrefを返す
        start=self
        for index,label in enumerate(full_label):
            if index>0:
                start=start.__dict__[label]
        return start
    
    def get_dependencies(self):
        from pkg_resources import get_distribution
        children=self.runAll()
        dependencies={}
        for child in children:
            try:
                m_name=child.__module__.split('.',1)[0]
                version=get_distribution(m_name).version
                dependencies[m_name]=version
            except:
                pass
        return dependencies
        
if __name__=='__main__':
    import sys
    import numpy as np
    import pandas as pd
    import sympy
    from coatl.Graph import MyGraphWindow
    app = QApplication([])
    g=MyGraphWindow()
    n=np.array([1,2,3])
    s=pd.Series()
    a=1
    b=[1,2,3]
    sym=sympy.Symbol('')
    mytree=MyRootTree()
    mytree.add(g)
    mytree.add(n)
    mytree.add(s)
    mytree.add(a)
    mytree.add(b)
    mytree.add(sym)
    print(mytree)
    print(mytree.get_dependencies())
    sys.exit(app.exec_())


# coding: utf-8

try:
    import setuptools
    from setuptools import setup, find_packages
except ImportError:
    print("Please install setuptools.")

import os
long_description = 'This package provides you functions for data analysis, namely, data browsing and graph plotting. It can save the products made by these functions to a file.'

setup(
    name  = 'coatl',
    version = '0.0',
    description = 'For data analysis',
    long_description = long_description,
    license = 'MIT',
    author = 'Yuki Arai',
    author_email = 'threemeninaboat3247@gmail.com',
    url = 'https://github.com/threemeninaboat3247/coatl',
    keywords = 'data analysis',
    packages = find_packages(),
    install_requires = ['numpy','matplotlib','pandas'],
    classifiers = [
      'Programming Language :: Python',
      'Intended Audience :: Science/Research',
      'License :: OSI Approved :: MIT License'
    ]
)



# -*- coding: utf-8 -*-
"""
Created on Sat Jan 21 02:40:20 2017

@author: Yuki
"""

import sys
import numpy as np
from scipy.optimize import curve_fit
from scipy.optimize import minimize
from coatl.Graph import MyGraphWindow

##############################################
#differential
##############################################

def differentiate(xs,ys,number=2):
    '''Neumerical diffferentiation by the linear regression.'''
    import numpy as np
    xs=np.array(xs)
    ys=np.array(ys)
    array=np.c_[xs,ys]
    array=array[array[:,0].argsort()] #xの昇順に並べ替え
    
    sort_xs=array[:,0]
    sort_ys=array[:,1]
    
    def least_square(sort_xs,sort_ys):
        #translate the coordinate system to avoid a cancellation of digits.
        n=len(sort_xs)
        xoffset=sort_xs[0]*np.ones(n)
        yoffset=sort_ys[0]*np.ones(n)
        sort_xs=sort_xs-xoffset
        sort_ys=sort_ys-yoffset

        x_ave=sort_xs.sum()/n
        y_ave=sort_ys.sum()/n
        xy_ave=np.dot(sort_xs,sort_ys)/n
        xx_ave=np.dot(sort_xs,sort_xs)/n

        a=(xy_ave-x_ave*y_ave)/(xx_ave-x_ave**2)
        #b=y_ave-a*x_ave
        #b=b+yoffset+a*xoffset #inverse to the original coordinate system
        return a
    
    df=np.ones(len(sort_xs))
    
    #左端number個の点は左側の点が不足するので処理を分ける
    for i in range(0,number):
        nearestx=sort_xs[0:i+number+1]
        nearesty=sort_ys[0:i+number+1]
        df[i]=least_square(nearestx,nearesty)
            
    for i in range(number,len(sort_xs)-number):
        nearestx=sort_xs[i-number:i+number+1]
        nearesty=sort_ys[i-number:i+number+1]
        df[i]=least_square(nearestx,nearesty)
        
    #右端number個の点は右側の点が不足するので処理を分ける
    for i in range(len(sort_xs)-number,len(sort_xs)):
        nearestx=sort_xs[i-number:]
        nearesty=sort_ys[i-number:]
        df[i]=least_square(nearestx,nearesty)
    
    return {'df':df,'sorted_xs':sort_xs,'sorted_ys':sort_ys}

##########################################################
#x,y方向にerrorが存在する場合の最小自乗法によるフィッティング関係
##########################################################    
class MyFitting():
    @classmethod
    def fit(cls,x,y,func,xlist,init_values,x_e=None,y_e=None):
        '''データ点にたいして指定された関数でfittingを行う グラフにデータ点とfitting結果をplotして結果をreturn'''
        #x,y:list
        #func:関数(params,x) params:関数パラメータ x:引数
        #xlist:定義域 list
        #init_values:関数パラメータの初期値
        #x_e,y_e標準偏差のlist
        if (x_e==None) or (y_e==None): #標準偏差が指定されない場合は全て同じ重みで計算
            print('use scipy.optimize.curve_fit')
        else:
            #MyDeviationPointsを作成
            points=[MyPoint(x,y) for x,y in zip(x,y)]
            dev_points=MyDeviationPoints(points,x_e,y_e)
            
            def handleFunc(params): #minimizeで最小化するべき評価値を返す
                ylist=[func(params,x) for x in xlist]
                points=[MyPoint(x,y) for x,y in zip(xlist,ylist)]
                curve=MyCurve(points)
                return dev_points.value(curve)
                
            result=minimize(handleFunc,init_values,method="nelder-mead")
            params_result=result.x
            
            def func_result(x): #fitting結果の関数
                return func(params_result,x)
                
            g=MyGraphWindow()
            g.plot(xlist,[func_result(x) for x in xlist])
            ax=g.fig.get_axes()[0]
            ax.errorbar(x,y,xerr=x_e,yerr=y_e,fmt='ro',markersize=4,ecolor='g')
            return {'params result':params_result,'function result':func_result,'graph':g}

class MyDeviationPoints():
    '''標準偏差付きのデータ点を保持するクラス'''
    def __init__(self,points,xerrors,yerrors):
        '''points:list of MyPoint ,xerrors:list,yerrors:list'''
        self.points=points
        self.xerrors=xerrors
        self.yerrors=yerrors
        self.weights=[1/(xerror**2+yerror**2) for xerror,yerror in zip(self.xerrors,self.yerrors)]
    
    @classmethod
    def distance(cls,point,curve):
        '''点とカーブを与えるとその距離を返す'''
        return curve.distance(point)
        
    def value(self,curve):
        '''カーブを与えると評価値を返す'''
        value=0.0
        for mypoint,weight in zip(self.points,self.weights):
            d=self.distance(mypoint,curve)
            value=value+weight*(d**2)
        return value    

class MyPoint():
    '''2Dの点を表すクラス'''
    def __init__(self,x,y):
        self.x=x
        self.y=y

    def distance(self,point):
        '''他点を与えるとそれとの距離を返す'''
        start=np.array([self.x,self.y])
        end=np.array([point.x,point.y])
        d=np.linalg.norm(end-start)
        return d
        
class MyCurve():
    '''曲線を表すクラス'''
    def __init__(self,points):
        self.points=points
        
    def distance(self,point):
        '''点を与えるとそれとの距離を返す　曲線が有限個の点で表現されていることに注意'''
        d=float('inf')
        for mypoint in self.points:
            temp=mypoint.distance(point)
            if temp<d:
                d=temp
        return d




# -*- coding: utf-8 -*-
"""
Created on Wed Apr 12 04:16:37 2017

@author: Yuki
"""
import matplotlib.pyplot as plt
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5 import NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure
from matplotlib.lines import Line2D
from matplotlib.widgets import LassoSelector
from matplotlib.path import Path

from PyQt5.QtCore import QRect,pyqtSignal,Qt,QTimer
from PyQt5.QtWidgets import QWidget,QHBoxLayout,QVBoxLayout,QLabel,QTextEdit,QSizePolicy,QPushButton,\
                            QMainWindow,QAction,QApplication
from PyQt5.QtGui import QColor

import numpy as np

#Default graph setting
plt.rcParams['font.family'] ='sans-serif'
plt.rcParams['xtick.direction'] = 'in'
plt.rcParams['ytick.direction'] = 'in'
plt.rcParams['xtick.top']=True
plt.rcParams['ytick.right']=True
plt.rcParams['xtick.major.width'] = 1.0
plt.rcParams['ytick.major.width'] = 1.0
plt.rcParams['font.size'] = 14 #フォントの大きさ
plt.rcParams['axes.linewidth'] = 1.0# 軸の線幅edge linewidth。囲みの太さ

##########################################################
#グラフウィンドウ
##########################################################
class MyPointer(QWidget):
    '''グラフにドロップできるポインタ　ドロップされると点にsnapしてx,yに点の座標を表示する'''
    COLOR='yellow'
    ALPHA=0.5
    SIZE=18
    def __init__(self):
        super().__init__()
        row=QHBoxLayout()
        xlabel=QLabel('X:')
        self.xEdit=QTextEdit()
        self.xEdit.setReadOnly(True)
        self.xEdit.setFixedHeight(30)
        self.xEdit.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        ylabel=QLabel('Y:')
        self.yEdit=QTextEdit()
        self.yEdit.setReadOnly(True)
        self.yEdit.setFixedHeight(30)
        self.yEdit.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        button=QPushButton('✖')
        button.pressed.connect(self.reset)
        row.addWidget(xlabel)
        row.addWidget(self.xEdit)
        row.addWidget(ylabel)
        row.addWidget(self.yEdit)
        row.addWidget(button)
        row.addStretch(1)
        
        self.setLayout(row)
        
    def reset(self):
        try:
            self.point.remove()
            self.xEdit.setText(None)
            self.yEdit.setText(None)
            self.canvas.draw()
        except:
            pass
        
    def setXY(self,event):
        if not len(event.ind):
            return True
        self.ax=event.artist.axes
        try: #すでにpointがあるとき
            if self.point.axes==self.ax: #同じaxにpointがあるとき
                pass
            else: #前と違うaxでクリックされた時
                self.point.remove()
                self.point,=self.ax.plot([], [], 'o', ms=self.SIZE, alpha=self.ALPHA,color=self.COLOR, visible=False)
        except: #初めてクリックした時
            self.point,=self.ax.plot([], [], 'o', ms=self.SIZE, alpha=self.ALPHA,color=self.COLOR, visible=False)
        ind = event.ind[0]
        x = event.artist.get_xdata()[ind]
        y = event.artist.get_ydata()[ind]
        self.target=event.artist
        self.ind=ind
        clicked = (x, y)
        
        self.point.set_visible(True)
        self.point.set_data(clicked)
        self.canvas.draw()
        self.xEdit.setText(str(x))
        self.yEdit.setText(str(y))
        
    def startMove(self,event):
        try:
            self.timer.stop()
        except:
            pass
        self.timer=QTimer(self)
        if event.key == 'right':
            self.timer.timeout.connect(self.moveRight)
        elif event.key=='left':
            self.timer.timeout.connect(self.moveLeft)
        self.timer.start(70)
        
    def stopMove(self,event):
        self.timer.stop()
        
    def moveRight(self):
        try:
            self.ind+=1
            x=self.target.get_xdata()[self.ind]
            y=self.target.get_ydata()[self.ind]
            self.point.set_data(x,y)
            self.canvas.draw()
            self.xEdit.setText(str(x))
            self.yEdit.setText(str(y))
        except:
            pass
        
    def moveLeft(self):
        try:
            self.ind-=1
            x=self.target.get_xdata()[self.ind]
            y=self.target.get_ydata()[self.ind]
            self.point.set_data(x,y)
            self.canvas.draw()
            self.xEdit.setText(str(x))
            self.yEdit.setText(str(y))
        except:
            pass
        
    def activate(self,canvas):
        self.show()
        self.canvas=canvas
        self.pick=self.canvas.mpl_connect('pick_event',self.setXY)
        self.key_press=self.canvas.mpl_connect('key_press_event',self.startMove)
        self.key_release=self.canvas.mpl_connect('key_release_event',self.stopMove)
        self.canvas.setFocusPolicy(Qt.ClickFocus )
        self.canvas.setFocus()
        
    def deactivate(self):
        self.hide()
        self.reset()
        try:
            self.canvas.mpl_disconnect(self.pick)
            self.canvas.mpl_disconnect(self.key_press)
            self.canvas.mpl_disconnect(self.key_release)
        except:
            pass
        
def geneMyGraph(fig,geometry,title):
    return MyGraphWindow(fig,geometry,title)
    
class MyColors():
    '''Default color list for MyGraphWindow.'''
    def __init__(self):
        self.index=0
        self.colors=['r','g','b','c','m','y','k']
        self.length=len(self.colors)

    def get_color(self):
        color=self.colors[self.index]
        self._incre_index()
        return color

    def _incre_index(self):
        self.index=(self.index+1)%self.length
        
class MyGraphWindow(QMainWindow):
    ALPHA_OTHER=0.2 #lasso非選択データの透明度
    PICKER=5 #lineをクリックしたときに取得できるための領域
    mycolors=MyColors() #memory the color to use for self.plot method.
    def __init__(self,fig=None,geo=None,title='temp'):
        super().__init__()
        #mutableなオブジェクトをデフォルトに指定するとまずいのでこのように書く
        if fig==None:
            self.fig=Figure(figsize=(6,6), dpi=100)
        else:
            self.fig=fig
        if geo==None:
            self.geo=QRect(30,30,500, 500)
        else:
            self.geo=geo
            
        self.canvas=FigureCanvas(self.fig)
        self.canvas.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.navi_toolbar = NavigationToolbar(self.canvas, self)
        self.lassoAction=QAction('lasso',self.canvas)
        self.pointerAction=QAction('pointer',self.canvas)
        self.lineAction=QAction('line',self.canvas)
        self.lassoAction.setCheckable(True)
        self.pointerAction.setCheckable(True)
        self.lineAction.setCheckable(True)
        self.navi_toolbar.addAction(self.lassoAction)
        self.navi_toolbar.addAction(self.pointerAction)
        self.navi_toolbar.addAction(self.lineAction)
        self.lassoAction.triggered.connect(self.lassoTriggered)
        self.pointerAction.triggered.connect(self.pointerTriggered)
        self.lineAction.triggered.connect(self.lineTriggered)
        
        self.lassoTarget=None #選択対象 lasso時はcilickでalphaを下げる
        self.selectedLine=None #選択中の点をself.lassoTargetの上にプロットする
        self.pre_alpha=1 #alphaを下げる前の値
        self.pre_settings=None #markerの色とサイズselectedLineをplotする時に使う
        
        #setTargetで指定したlineの点とそこからlassoで選ばれたindexを保存
        self.xys=None #numpy.array 列数2の行列
        self.selectedIndices=set()
        self.unselectedIndices=set()
        
        #クリックされたlineを記憶しておいてgetLineで参照を渡す
        self.line=None
        self.pre_width=0 #太くした後元に戻すのに使う
        
        #点の座標を表示する
        self.pointer=MyPointer()
        self.pointer.deactivate()
        
        self.vbl = QVBoxLayout()
        self.vbl.addWidget(self.navi_toolbar)
        self.vbl.addWidget( self.canvas )
        self.vbl.addWidget(self.pointer)
        
        self.frontPanel=QWidget()
        self.frontPanel.setLayout(self.vbl)
        self.setCentralWidget(self.frontPanel)
        
        self.setGeometry(self.geo)
        self.setWindowTitle(title)
        self.show()
        
    def __reduce_ex__(self, proto):
        #pickleのためのメソッド QMainWindowはそのままpickleできないことに注意
        return geneMyGraph, (self.fig,self.geometry(),self.windowTitle())
        
    def getSelected(self):
        points=self.xys[list(self.selectedIndices)]
        return points
        
    def getUnSelected(self):
        points=self.xys[list(self.unselectedIndices)]
        return points
        
    def getLine(self):
        return self.line
        
    def plot(self,*args,**kwargs):
        #pickerはlineの選択に使うので必要
        kwargs=self._addPickerToValue(kwargs)
        kwargs=self._addColorToValue(kwargs)

        try: #axが指定されていたらそのaxに対して描画を行う axが無いまたはaxオブジェクト以外が指定されていると引数にaxが無い状態で次
            ax=kwargs.pop('ax')
            ax.plot(*args,**kwargs)
        except: #無ければself.figの最初のaxに描画　それもなければ新たにaxを追加して描画
            try:
                ax=self.fig.get_axes()[0]
                ax.plot(*args,**kwargs)
            except:
                ax=self.fig.add_subplot(111)
                ax.plot(*args,**kwargs)
        self.canvas.draw()
        
    def _addPickerToValue(self,kwargs):
        #pickerが無いとlassoで選択できないのでユーザーが引数にpickerを指定しなかった場合はpickerを加える
        if 'picker' in kwargs.keys():
            return kwargs
        else:
            kwargs['picker']=self.PICKER
            return kwargs
            
    def _addColorToValue(self,kwargs):
        #If color is in the kwargs use it. Otherwise get a color from self.mycolors and use it.
        if not 'color' in kwargs.keys():
            color=self.mycolors.get_color()
            kwargs['color']=color
        return kwargs
                
    def lassoTriggered(self):
        if self.lassoAction.isChecked()==True: #pointer Off lasso　On
            if self.pointerAction.isChecked()==True:
                self.pointerAction.setChecked(False)
                self.pointer.deactivate()
            if self.lineAction.isChecked()==True:
                self.lineAction.setChecked(False)
                self.removeLine()
            self.setLasso()
        else:
            self.removeLasso()
            
    def pointerTriggered(self):
        if self.pointerAction.isChecked()==True: #pointer on lasso　off
            if self.lassoAction.isChecked()==True:
                self.lassoAction.setChecked(False)
                self.removeLasso()
            if self.lineAction.isChecked()==True:
                self.lineAction.setChecked(False)
                self.removeLine()
            self.pointer.activate(self.canvas)
        else:
            self.pointer.deactivate()
            
    def lineTriggered(self):
        if self.lineAction.isChecked()==True: #他をoff
            if self.lassoAction.isChecked()==True:
                self.lassoAction.setChecked(False)
                self.removeLasso()
            if self.pointerAction.isChecked()==True:
                self.pointerAction.setChecked(False)
                self.pointer.deactivate()
            self.setLine()
        else:
            self.removeLine()
            
    def setLine(self):
        self.linepick=self.canvas.mpl_connect('pick_event',self.highlightLine)
        
    def removeLine(self):
        try:
            self.canvas.mpl_disconnect(self.linepick)
            if not self.line==None: #選択されていたものを元に戻す
                self.line.set_linewidth(self.pre_width)
                self.canvas.draw()
            self.line=None
            self.pre_width=0
        except:
            print('line is still not sellected')
        
    def highlightLine(self,event):
        #選択中のlineの輪郭を太くする
        if not len(event.ind):
            return True
        if not self.line==None: #初回のclickでない場合は先に選択されていたものを元に戻す
            self.line.set_linewidth(self.pre_width)
        #plotの設定を保存
        self.line=event.artist
        self.pre_width=self.line.get_linewidth()
        
        #alphaの調整とselectedLineの描画 まだinvisible　データの取得
        self.line.set_linewidth(self.pre_width*5)
        self.canvas.draw()
            
    def setLasso(self):
        self.pick=self.canvas.mpl_connect('pick_event',self.setTarget)
        
    def removeLasso(self):
        try:
            self.canvas.mpl_disconnect(self.pick)
            del self.lasso
            if not self.lassoTarget==None: #先に選択されていたものを元に戻す
                self.lassoTarget.set_alpha(self.pre_alpha)
                self.selectedLine.remove()
                self.canvas.draw()
            #多分必要ないが念のために初期化
            self.lassoTarget=None
            self.selectedLine=None
            self.pre_alpha=1
            self.pre_settings=None
            self.xys=None
            self.selectedIndices=set()
            self.unselectedIndices=set()
        except:
            print('Lasso is still not setted')
            
    def setTarget(self,event):
        #lassoの対象をclickから決定する
        if not len(event.ind):
            return True
        if not self.lassoTarget==None: #初回のclickでない場合は先に選択されていたものを元に戻す
            self.lassoTarget.set_alpha(self.pre_alpha)
            try: #あったら掃除 多分いつもある
                self.selectedLine.remove()
            except:
                pass
        self.selectedIndices=set()
        self.unselectedIndices=set()
        
        #plotの設定を保存
        self.lassoTarget=event.artist
        self.pre_alpha=self.lassoTarget.get_alpha()
        self.pre_settings=self.getSettings(self.lassoTarget)
        
        #alphaの調整とselectedLineの描画 まだinvisible　データの取得
        self.lassoTarget.set_alpha(self.ALPHA_OTHER)
        self.selectedLine,=self.lassoTarget.axes.plot([],[],visible=False,linestyle='None',**self.pre_settings)
        self.lasso=LassoSelector(self.lassoTarget.axes, onselect=self.onselect)
        self.xys=np.array([[x,y] for x,y in zip(self.lassoTarget.get_xdata(),self.lassoTarget.get_ydata())])
        self.canvas.draw()
        
    def getSettings(self,line):
        settings={'color':line.get_color(),'marker':line.get_marker(),'markersize':line.get_markersize()}
        return settings

    def onselect(self, verts):
        path = Path(verts)
        try:
            selected_ind = set(np.nonzero([path.contains_point(xy) for xy in self.xys])[0])
            all_ind=set(range(self.xys.shape[0]))
            unselected_ind=all_ind-selected_ind
        except:
            print('canot determin selected points')
        #選択される度にselcectedlineに追加
        self.selectedIndices=self.selectedIndices | selected_ind
        self.unselectedIndices=self.unselectedIndices | unselected_ind
        self.selectedLine.set_visible(True)
        points=self.xys[list(self.selectedIndices)]
        self.selectedLine.set_data(points[:,0],points[:,1])
        self.canvas.draw()
        
if __name__=='__main__':
    import sys
    app = QApplication( sys.argv )
    g=MyGraphWindow()
    x=np.arange(0,1000,0.1)
    y=np.random.rand(10000)
    z=np.random.rand(10000)+10
    g.plot(x,y,marker='o',markersize=20)
    g.plot(x,z,marker='o',picker=20)
    sys.exit(app.exec_())